"""
Declare credit retirement DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from __future__ import annotations
import datetime as dt
import typing

from sqlalchemy import ForeignKey, Index, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from sc_audit.db_schema.base import ScBase, intpk, strkey
from sc_audit.db_schema.impact_project import VcsProject

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import RetirementFromBlock, SinkStatus


InstrumentType = typing.Literal['VCU']


class Retirement(ScBase):
    __tablename__ = "retirements"

    certificate_id: Mapped[intpk]
    vcu_amount: Mapped[int]
    serial_number: Mapped[str] = mapped_column(String(128))
    retirement_date: Mapped[dt.date]
    retirement_beneficiary: Mapped[strkey]
    retirement_details: Mapped[str]
    vcs_project_id: Mapped[int] = mapped_column(ForeignKey("vcs_projects.id"))
    vcs_project: Mapped[VcsProject] = relationship(init=False, repr=False)
    issuance_date: Mapped[dt.date]
    instrument_type: Mapped[InstrumentType]
    vintage_start: Mapped[dt.date]
    vintage_end: Mapped[dt.date]
    total_vintage_quantity: Mapped[int]

    retired_from: Mapped[list[RetirementFromBlock]] = relationship(
        init=False, repr=False, 
        back_populates='retirement',
    )
    sink_statuses: Mapped[list[SinkStatus]] = relationship(
        init=False, repr=False, 
        back_populates='retirement',
    )

    @property
    def tx_hashes_from_details(self) -> list[str]:
        tag, *tx_hashes = self.retirement_details.split()
        assert tag.startswith("stellarcarbon")
        return tx_hashes


idx_retirement_date = Index("idx_retirement_date", Retirement.retirement_date.desc())
idx_retirement_beneficiary = Index("idx_retirement_beneficiary", Retirement.retirement_beneficiary)



import os
from pathlib import Path

from stellar_sdk import Asset

def get_default_db_path(db_name="sc-audit.sqlite3") -> Path:
    """
    Return the default location of the SQLite DB, as determined by:

    if the current working dir is a descendent of the home dir:
        use the current working dir
    else:
        use a subdir of the home dir and ensure it exists
    """
    home_dir = Path.home()
    work_dir = Path.cwd()

    if home_dir in work_dir.resolve().parents:
        if work_dir.name == "sc_audit":
            return work_dir.parent / db_name
        
        return work_dir / db_name
    else:
        db_dir = home_dir / "sc-audit"
        db_dir.mkdir(exist_ok=True)
        return db_dir / db_name


DBAPI_URL: str = os.environ.get('SC_DBAPI_URL', f"sqlite+pysqlite:///{get_default_db_path()}")

VERRA_ASSET_SEARCH_URL = "https://registry.verra.org/uiapi/asset/asset/search"
VERRA_ASSET_SEARCH_TIMEOUT = 8  # seconds
VERRA_REPORT_URL = "https://registry.verra.org/mymodule/rpt/myRpt.asp"

HORIZON_URL: str = os.environ.get('SC_HORIZON_URL',"https://horizon.stellar.org")
CARBON_ISSUER_PUB = "GCBOATLWKXACOWKRRWORARDI2HFDSYPALMTS23YBZKHOB6XLW6CARBON"
CARBON_DISTRIB_PUB = "GABXJLJDNWRIGCVQPPC2FJ2NORUJBGLNOBYXKW7WQD3CUZS6YTFCS4ZP"
SINK_ISSUER_PUB = "GC7CDWMCWNCY7JYUW5UBEOLNBSTNDKKZSFTHKGZNPPSOXLFYFX3CSINK"
CARBON_ASSET = Asset("CARBON", CARBON_ISSUER_PUB)
SINK_ASSET = Asset("CarbonSINK", SINK_ISSUER_PUB)
FIRST_SINK_CURSOR = 164821723627237376
FIRST_MINT_CURSOR = 164806777139924992
FIRST_DIST_CURSOR = 164810659791396865



"""
SQLAlchemy Declarative Dataclass setup.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from decimal import Decimal
from typing import Annotated

from sqlalchemy import Dialect, String
import sqlalchemy.types as types
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass


class HexBinary(types.TypeDecorator):
    """
    Converts between fixed-length bytes and their hexadecimal string representations.
    """
    impl = types.BINARY
    cache_ok = True

    def process_bind_param(self, value: str, dialect: Dialect) -> bytes:
        return bytes.fromhex(value)
    
    def process_result_value(self, value: bytes, dialect: Dialect) -> str:
        if value is None:
            # this can happen during joins or eager loading
            return ""
        
        return value.hex()


intpk = Annotated[int, mapped_column(primary_key=True)]
hashpk = Annotated[str, mapped_column(HexBinary(length=32), primary_key=True)]
strkey = Annotated[str, mapped_column(String(56))]
kgdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=3))]
stroopdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=7))]


class ScBase(MappedAsDataclass, DeclarativeBase):
    pass



"""
View sinking transactions and their retirement status.

Transactions can be filtered by account, by date, and by their retirement status.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt

import pandas as pd
from sqlalchemy import Select, or_, select
from sqlalchemy.orm import contains_eager

from sc_audit.db_schema.association import SinkStatus
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


def view_sinking_txs(
        for_funder: str | None = None,
        for_recipient: str | None = None, 
        from_date: dt.date | None = None,
        before_date: dt.date | None = None,
        finalized: bool | None = None,
) -> pd.DataFrame:
    stx_query = construct_stx_query(for_funder, for_recipient, from_date, before_date, finalized)
    with Session.begin() as session:
        stx_records = session.scalars(stx_query).unique().all()
        txdf = pd.DataFrame.from_records(stx.as_dict() for stx in stx_records)

    return txdf
    
def construct_stx_query(
        for_funder: str | None,
        for_recipient: str | None, 
        from_date: dt.date | None,
        before_date: dt.date | None,
        finalized: bool | None,
) -> Select[tuple[SinkingTx]]:
    q_txs = select(SinkingTx).outerjoin(SinkStatus).options(contains_eager(SinkingTx.statuses))
    if for_funder:
        q_txs = q_txs.where(SinkingTx.funder == for_funder)

    if for_recipient:
        q_txs = q_txs.where(SinkingTx.recipient == for_recipient)

    if from_date:
        from_dt = dt.datetime(from_date.year, from_date.month, from_date.day)
        q_txs = q_txs.where(SinkingTx.created_at >= from_dt)

    if before_date:
        before_dt = dt.datetime(before_date.year, before_date.month, before_date.day)
        q_txs = q_txs.where(SinkingTx.created_at < before_dt)

    if finalized is False:
        # not_ and in_ do not work here because NULL and false are treated differently
        q_txs = q_txs.where(or_(SinkStatus.finalized == None, SinkStatus.finalized == False))
    elif finalized is True:
        q_txs = q_txs.where(SinkStatus.finalized == True)

    return q_txs



"""
DB connection management.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import sys
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sc_audit.constants import DBAPI_URL

print(f"Connecting to database {DBAPI_URL}...", file=sys.stderr)
engine = create_engine(DBAPI_URL)
Session = sessionmaker(engine)
